import torch.optim.lr_scheduler as scheduler


# class ReduceLrOnPlateau:
#     def __init__(self, optimizer, metric_name, mode, factor, patience, threshold, threshold_mode, cooldown, min_lr, epsilon):
#         self.metric_name = metric_name
#         self.scheduler = scheduler.ReduceLROnPlateau(
#             optimizer,
#             mode=mode,
#             factor=factor,
#             patience=patience,
#             threshold=threshold,
#             threshold_mode=threshold_mode,
#             cooldown=cooldown,
#             min_lr=min_lr,
#             eps=epsilon
#         )
#
#     def get_lr(self):
#         return [p['lr'] for p in self.scheduler.optimizer.param_groups]
#
#     def pre_epoch_step(self):
#         pass
#
#     def post_epoch_step(self, metrics):
#         self.scheduler.step(metrics[self.metric_name])
#
#
# def create(metric_name, mode='min', factor=0.1, patience=10, threshold=1e-4, threshold_mode='rel', cooldown=10,
#            min_lr=0, epsilon=1e-8):
#     """ Create a scheduler that lowers the LR on metric plateau """
#     def scheduler_fn(optimizer):
#         return ReduceLrOnPlateau(optimizer, metric_name, mode, factor, patience, threshold, threshold_mode, cooldown, min_lr, epsilon)
#
#     return scheduler_fn




"""
Code based loosely on implementation:
https://github.com/keras-team/keras/blob/master/examples/mnist_cnn.py

Under MIT license.
"""
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F

from vel.api.base import SupervisedModel, ModelFactory
from vel.metrics.loss_metric import Loss
from vel.metrics.accuracy import Accuracy


class Net(SupervisedModel):
    """
    A simple MNIST classification model.

    Conv 3x3 - 32
    Conv 3x3 - 64
    MaxPool 2x2
    Dropout 0.25
    Flatten
    Dense - 128
    Dense - output (softmax)
    """

    @staticmethod
    def _weight_initializer(tensor):
        init.xavier_uniform_(tensor.weight, gain=init.calculate_gain('relu'))
        init.constant_(tensor.bias, 0.0)

    def __init__(self, img_rows, img_cols, img_channels, num_classes):
        super(Net, self).__init__()

        self.flattened_size = (img_rows - 4) // 2 * (img_cols - 4) // 2 * 64

        self.conv1 = nn.Conv2d(in_channels=img_channels, out_channels=32, kernel_size=(3, 3))
        self.conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(3, 3))

        self.dropout1 = nn.Dropout2d(p=0.25)

        self.fc1 = nn.Linear(self.flattened_size, 128)
        self.dropout2 = nn.Dropout(p=0.5)
        self.fc2 = nn.Linear(128, num_classes)

    def reset_weights(self):
        self._weight_initializer(self.conv1)
        self._weight_initializer(self.conv2)
        self._weight_initializer(self.fc1)
        self._weight_initializer(self.fc2)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, kernel_size=(2, 2))
        x = self.dropout1(x)
        x = x.view(-1, self.flattened_size)
        x = F.relu(self.fc1(x))
        x = self.dropout2(x)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)

    def loss_value(self, x_data, y_true, y_pred):
        """ Calculate a value of loss function """
        return F.nll_loss(y_pred, y_true)

    def metrics(self):
        """ Set of metrics for this model """
        return [Loss(), Accuracy()]


def create(img_rows, img_cols, img_channels, num_classes):
    """ Create the model matching specified image dimensions """
    def instantiate(**_):
        return Net(img_rows, img_cols, img_channels, num_classes)

    return ModelFactory.generic(instantiate)



import numpy as np
import tqdm

import torch
import torch.nn.functional as F
import torch.distributions as dist

from vel.api import TrainingInfo


class GenerateTextCommand:
    """ Generate text using RNN model """

    def __init__(self, model_config, model_factory, source, storage, start_letter, length, temperature):
        self.model_config = model_config
        self.model_factory = model_factory
        self.source = source
        self.storage = storage
        self.start_letter = start_letter
        self.length = length
        self.temperature = temperature

    @torch.no_grad()
    def run(self):
        device = torch.device(self.model_config.device)
        model = self.model_factory.instantiate().to(device)

        start_epoch = self.storage.last_epoch_idx()

        training_info = TrainingInfo(
            start_epoch_idx=start_epoch,
            run_name=self.model_config.run_name,
        )

        self.storage.resume(training_info, model)

        model.eval()

        current_char = self.start_letter
        current_char_encoded = self.source.encode_character(self.start_letter)

        generated_text = [current_char]

        state = model.initial_state(1).to(device)

        char_tensor = torch.from_numpy(np.array([current_char_encoded])).view(1, 1).to(device)

        for _ in tqdm.trange(self.length):
            prob_logits, state = model.forward_state(char_tensor, state)

            # Apply temperature to the logits
            prob_logits = F.log_softmax(prob_logits.view(-1).div(self.temperature), dim=0)

            distribution = dist.Categorical(logits=prob_logits)

            char_tensor = distribution.sample().view(1, 1)
            current_char_encoded = char_tensor.item()

            if current_char_encoded == 0:
                # End of sequence marker
                break

            current_char = self.source.decode_character(current_char_encoded)

            generated_text.append(current_char)

        print("============================ START GENERATED TEXT ================================================")
        print(''.join(generated_text))
        print("============================ END GENERATED TEXT ================================================")


def create(model_config, model, source, storage, start_letter, length, temperature):
    return GenerateTextCommand(model_config, model, source, storage, start_letter, length, temperature)



import gym
import itertools as it
import torch

from vel.api.base import LinearBackboneModel, Model, ModelFactory
from vel.rl.api import Rollout, Evaluator
from vel.rl.modules.deterministic_action_head import DeterministicActionHead
from vel.rl.modules.deterministic_critic_head import DeterministicCriticHead


class DeterministicPolicyEvaluator(Evaluator):
    """ Evaluator for DeterministicPolicyModel """

    def __init__(self, model: 'DeterministicPolicyModel', rollout: Rollout):
        super().__init__(rollout)

        self.model = model

    @Evaluator.provides('model:estimated_values_next')
    def model_estimated_values_next(self):
        """ Estimate state-value of the transition next state """
        observations = self.get('rollout:observations_next')
        action, value = self.model(observations)
        return value

    @Evaluator.provides('model:actions')
    def model_actions(self):
        """ Estimate state-value of the transition next state """
        observations = self.get('rollout:observations')
        model_action = self.model.action(observations)
        return model_action

    @Evaluator.provides('model:model_action:q')
    def model_model_action_q(self):
        observations = self.get('rollout:observations')
        model_actions = self.get('model:actions')
        return self.model.value(observations, model_actions)

    @Evaluator.provides('model:action:q')
    def model_action_q(self):
        observations = self.get('rollout:observations')
        rollout_actions = self.get('rollout:actions')
        return self.model.value(observations, rollout_actions)


class DeterministicPolicyModel(Model):
    """ Deterministic Policy Gradient - model """

    def __init__(self, policy_backbone: LinearBackboneModel, value_backbone: LinearBackboneModel,
                 action_space: gym.Space):
        super().__init__()

        self.policy_backbone = policy_backbone
        self.value_backbone = value_backbone

        self.action_head = DeterministicActionHead(self.policy_backbone.output_dim, action_space)
        self.critic_head = DeterministicCriticHead(self.value_backbone.output_dim)

    def reset_weights(self):
        """ Initialize properly model weights """
        self.policy_backbone.reset_weights()
        self.value_backbone.reset_weights()
        self.action_head.reset_weights()
        self.critic_head.reset_weights()

    def forward(self, observations, input_actions=None):
        """ Calculate model outputs """
        observations = observations.float()

        if input_actions is not None:
            actions = input_actions

            value_input = torch.cat([observations, actions], dim=1)
            value_hidden = self.value_backbone(value_input)

            values = self.critic_head(value_hidden)
        else:
            policy_hidden = self.policy_backbone(observations)
            actions = self.action_head(policy_hidden)

            value_input = torch.cat([observations, actions], dim=1)
            value_hidden = self.value_backbone(value_input)

            values = self.critic_head(value_hidden)

        return actions, values

    def policy_parameters(self):
        """ Parameters of policy """
        return it.chain(self.policy_backbone.parameters(), self.action_head.parameters())

    def value_parameters(self):
        """ Parameters of policy """
        return it.chain(self.value_backbone.parameters(), self.critic_head.parameters())

    def get_layer_groups(self):
        """ Return layers grouped """
        return [
            [self.policy_backbone, self.action_head],
            [self.value_backbone, [y for (x, y) in self.critic_head.named_parameters() if x.endswith('bias')]],
            # OpenAI regularizes only weight on the last layer. I'm just replicating that
            [[y for (x, y) in self.critic_head.named_parameters() if x.endswith('weight')]]
        ]

    def step(self, observation):
        """ Select actions based on model's output """
        action, value = self(observation)

        return {
            'actions': action,
            'values': value
        }

    def value(self, observation, input_actions=None):
        """ Calculate value for given state """
        action, value = self(observation, input_actions)
        return value

    def action(self, observations):
        """ Calculate value for given state """
        observations = observations.float()
        policy_hidden = self.policy_backbone(observations)
        action = self.action_head(policy_hidden)
        return action

    def evaluate(self, rollout: Rollout) -> Evaluator:
        """ Evaluate model on a rollout """
        return DeterministicPolicyEvaluator(self, rollout)


class DeterministicPolicyModelFactory(ModelFactory):
    """ Factory  class for policy gradient models """
    def __init__(self, policy_backbone: ModelFactory, value_backbone: ModelFactory):
        self.policy_backbone = policy_backbone
        self.value_backbone = value_backbone

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        policy_backbone = self.policy_backbone.instantiate(**extra_args)
        value_backbone = self.value_backbone.instantiate(**extra_args)

        return DeterministicPolicyModel(
            policy_backbone=policy_backbone,
            value_backbone=value_backbone,
            action_space=extra_args['action_space'],
        )


def create(policy_backbone: ModelFactory, value_backbone: ModelFactory):
    """ Vel creation function """
    return DeterministicPolicyModelFactory(
        policy_backbone=policy_backbone, value_backbone=value_backbone
    )



import collections.abc as abc
import numpy as np
import pandas as pd
import typing

import torch

from vel.exceptions import VelException


class TrainingHistory:
    """
    Simple aggregator for the training history.

    An output of training storing scalar metrics in a pandas dataframe.
    """
    def __init__(self):
        self.data = []

    def add(self, epoch_result):
        """ Add a datapoint to the history """
        self.data.append(epoch_result)

    def frame(self):
        """ Return history dataframe """
        return pd.DataFrame(self.data).set_index('epoch_idx')


class TrainingInfo(abc.MutableMapping):
    """
    Information that need to persist through the whole training process

    Data dict is any extra information processes may want to store
    """

    def __init__(self, start_epoch_idx=0, run_name: typing.Optional[str]=None, metrics=None, callbacks=None):
        self.data_dict = {}

        self.start_epoch_idx = start_epoch_idx
        self.metrics = metrics if metrics is not None else []
        self.callbacks = callbacks if callbacks is not None else []
        self.run_name = run_name
        self.history = TrainingHistory()

        self.optimizer_initial_state = None

    def restore(self, hidden_state):
        """ Restore any state from checkpoint - currently not implemented but possible to do so in the future """
        for callback in self.callbacks:
            callback.load_state_dict(self, hidden_state)

        if 'optimizer' in hidden_state:
            self.optimizer_initial_state = hidden_state['optimizer']

    def initialize(self):
        """
        Runs for the first time a training process is started from scratch. Is guaranteed to be run only once
        for the training process. Will be run before all other callbacks.
        """
        for callback in self.callbacks:
            callback.on_initialization(self)

    def on_train_begin(self):
        """
        Beginning of a training process - is run every time a training process is started, even if it's restarted from
        a checkpoint.
        """
        for callback in self.callbacks:
            callback.on_train_begin(self)

    def on_train_end(self):
        """
        Finalize training process. Runs each time at the end of a training process.
        """
        for callback in self.callbacks:
            callback.on_train_end(self)

    def __getitem__(self, item):
        return self.data_dict[item]

    def __setitem__(self, key, value):
        self.data_dict[key] = value

    def __delitem__(self, key):
        del self.data_dict[key]

    def __iter__(self):
        return iter(self.data_dict)

    def __len__(self):
        return len(self.data_dict)

    def __contains__(self, item):
        return item in self.data_dict


class EpochResultAccumulator:
    """ Result of a single epoch of training -- accumulator """
    def __init__(self, global_epoch_idx, metrics):
        self.global_epoch_idx = global_epoch_idx
        self.metrics = metrics

        self.frozen_results = {}

        self._reset_metrics()
        self.metrics_by_name = {m.name: m for m in self.metrics}

    @torch.no_grad()
    def calculate(self, batch_info):
        """ Calculate metric values """
        for m in self.metrics:
            m.calculate(batch_info)

    def _reset_metrics(self):
        """ Internal API : reset state of metrics """
        for m in self.metrics:
            m.reset()

    def value(self):
        """ Return current value of the metrics """
        return {m.name: m.value() for m in self.metrics}

    def intermediate_value(self, metric):
        """ Return an intermediate (inter-epoch) value of a metric """
        if ':' in metric:
            metric_name = metric.split(':')[-1]
        else:
            metric_name = metric

        return self.metrics_by_name[metric_name].value()

    def freeze_results(self, name=None):
        new_results = self.value()

        if name is None:
            for key, value in new_results.items():
                self.frozen_results[key] = value
        else:
            for key, value in new_results.items():
                self.frozen_results[f'{name}:{key}'] = value

        self._reset_metrics()

    def result(self):
        """ Return the epoch result """
        final_result = {'epoch_idx': self.global_epoch_idx}

        for key, value in self.frozen_results.items():
            final_result[key] = value

        return final_result


class EpochInfo(abc.MutableMapping):
    """
    Information that need to persist through the single epoch.

    Global epoch index - number of epoch from start of training until now
    Local epoch index - number of epoch from start of current "phase" until now

    Data dict is any extra information processes may want to store
    """

    def __init__(self, training_info: TrainingInfo, global_epoch_idx: int, batches_per_epoch: int,
                 optimizer: torch.optim.Optimizer=None, local_epoch_idx: int = None, callbacks: list=None):
        self.training_info = training_info
        self.optimizer = optimizer
        self.batches_per_epoch = batches_per_epoch

        self.global_epoch_idx = global_epoch_idx

        if local_epoch_idx is None:
            self.local_epoch_idx = self.global_epoch_idx
        else:
            self.local_epoch_idx = local_epoch_idx

        self.result_accumulator = EpochResultAccumulator(self.global_epoch_idx, self.training_info.metrics)
        self._result = {}
        self.data_dict = {}

        if callbacks is None:
            self.callbacks = self.training_info.callbacks
        else:
            self.callbacks = callbacks

    def state_dict(self) -> dict:
        """ Calculate hidden state dictionary """
        hidden_state = {}

        if self.optimizer is not None:
            hidden_state['optimizer'] = self.optimizer.state_dict()

        for callback in self.callbacks:
            callback.write_state_dict(self.training_info, hidden_state)

        return hidden_state

    def on_epoch_begin(self) -> None:
        """ Initialize an epoch """
        for callback in self.callbacks:
            callback.on_epoch_begin(self)

    def on_epoch_end(self):
        """ Finish epoch processing """
        self.freeze_epoch_result()

        for callback in self.callbacks:
            callback.on_epoch_end(self)

        self.training_info.history.add(self.result)

    @property
    def metrics(self):
        """ Just forward metrics from training_info """
        return self.training_info.metrics

    def freeze_epoch_result(self):
        """ Calculate epoch 'metrics' result and store it in the internal variable """
        self._result = self.result_accumulator.result()

    @property
    def result(self) -> dict:
        """ Result of the epoch """
        if self._result is None:
            raise VelException("Result has not been frozen yet")
        else:
            return self._result

    def __repr__(self):
        return f"EpochInfo(global_epoch_idx={self.global_epoch_idx}, local_epoch_idx={self.local_epoch_idx})"

    def __getitem__(self, item):
        return self.data_dict[item]

    def __setitem__(self, key, value):
        self.data_dict[key] = value

    def __delitem__(self, key):
        del self.data_dict[key]

    def __iter__(self):
        return iter(self.data_dict)

    def __len__(self):
        return len(self.data_dict)

    def __contains__(self, item):
        return item in self.data_dict


class BatchInfo(abc.MutableMapping):
    """
    Information about current batch.

    Serves as a dictionary where all the information from the single run are accumulated
    """
    def __init__(self, epoch_info: EpochInfo, batch_number: int):
        self.epoch_info = epoch_info
        self.batch_number = batch_number
        self.data_dict = {}

    def on_batch_begin(self):
        """ Initialize batch processing """
        for callback in self.callbacks:
            callback.on_batch_begin(self)

    def on_batch_end(self):
        """ Finalize batch processing """
        for callback in self.callbacks:
            callback.on_batch_end(self)

        # Even with all the experience replay, we count the single rollout as a single batch
        self.epoch_info.result_accumulator.calculate(self)

    def on_validation_batch_begin(self):
        """ Initialize batch processing """
        for callback in self.callbacks:
            callback.on_validation_batch_begin(self)

    def on_validation_batch_end(self):
        """ Finalize batch processing """
        for callback in self.callbacks:
            callback.on_validation_batch_end(self)

        # Even with all the experience replay, we count the single rollout as a single batch
        self.epoch_info.result_accumulator.calculate(self)

    @property
    def aggregate_batch_number(self):
        return self.batch_number + self.epoch_info.batches_per_epoch * (self.epoch_info.global_epoch_idx - 1)

    @property
    def epoch_number(self):
        return self.epoch_info.global_epoch_idx

    @property
    def batches_per_epoch(self):
        return self.epoch_info.batches_per_epoch

    @property
    def local_epoch_number(self):
        return self.epoch_info.local_epoch_idx

    @property
    def optimizer(self):
        return self.epoch_info.optimizer

    @property
    def training_info(self):
        return self.epoch_info.training_info

    @property
    def callbacks(self):
        return self.epoch_info.callbacks

    def aggregate_key(self, aggregate_key):
        """ Aggregate values from key and put them into the top-level dictionary """
        aggregation = self.data_dict[aggregate_key]  # List of dictionaries of numpy arrays/scalars

        # Aggregate sub batch data
        data_dict_keys = {y for x in aggregation for y in x.keys()}

        for key in data_dict_keys:
            # Just average all the statistics from the loss function
            stacked = np.stack([d[key] for d in aggregation], axis=0)
            self.data_dict[key] = np.mean(stacked, axis=0)

    def drop_key(self, key):
        """ Remove key from dictionary """
        del self.data_dict[key]

    def __getitem__(self, item):
        return self.data_dict[item]

    def __setitem__(self, key, value):
        self.data_dict[key] = value

    def __delitem__(self, key):
        del self.data_dict[key]

    def __iter__(self):
        return iter(self.data_dict)

    def __len__(self):
        return len(self.data_dict)

    def __contains__(self, item):
        return item in self.data_dict

    def __repr__(self):
        return f"[BatchInfo epoch:{self.epoch_info.global_epoch_idx} batch:{self.batch_number}/{self.batches_per_epoch}]"



from .base_metric import BaseMetric
from .averaging_metric import AveragingMetric, AveragingNamedMetric, AveragingSupervisedMetric
from .value_metric import ValueMetric
from .summing_metric import SummingMetric, SummingNamedMetric



import torch.nn.utils


class AlgoBase:
    """ Base class for algo reinforcement calculations """

    def initialize(self, settings, model, environment, device):
        """ Initialize algo from reinforcer settings """
        pass

    def optimizer_step(self, batch_info, device, model, rollout):
        """ Single optimization step for a model """
        raise NotImplementedError

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return []

    def _clip_gradients(self, batch_result, model, max_grad_norm):
        """ Clip gradients to a given maximum length """
        if max_grad_norm is not None:
            grad_norm = torch.nn.utils.clip_grad_norm_(
                filter(lambda p: p.requires_grad, model.parameters()),
                max_norm=max_grad_norm
            )
        else:
            grad_norm = 0.0

        batch_result['grad_norm'] = grad_norm


class OptimizerAlgoBase(AlgoBase):
    """ RL algo that does a simple optimizer update """

    def __init__(self, max_grad_norm):
        self.max_grad_norm = max_grad_norm

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        raise NotImplementedError

    def post_optimization_step(self, batch_info, device, model, rollout):
        """ Steps to take after optimization has been done"""
        pass

    def optimizer_step(self, batch_info, device, model, rollout):
        """ Single optimization step for a model """
        batch_info.optimizer.zero_grad()

        batch_result = self.calculate_gradient(batch_info=batch_info, device=device, model=model, rollout=rollout)

        self._clip_gradients(batch_result, model, self.max_grad_norm)

        batch_info.optimizer.step(closure=None)

        self.post_optimization_step(batch_info, device, model, rollout)

        return batch_result

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return []



import numpy as np

from .base_metric import BaseMetric


class AveragingMetric(BaseMetric):
    """ Base class for metrics that simply calculate the average over the epoch """
    def __init__(self, name):
        super().__init__(name)

        self.storage = []

    def calculate(self, batch_info):
        """ Calculate value of a metric """
        value = self._value_function(batch_info)
        self.storage.append(value)

    def _value_function(self, batch_info):
        raise NotImplementedError

    def reset(self):
        """ Reset value of a metric """
        self.storage = []

    def value(self):
        """ Return current value for the metric """
        return float(np.mean(self.storage))


class AveragingNamedMetric(AveragingMetric):
    """ Super simple averaging metric that just takes a value from dictionary and averages it over samples """
    def __init__(self, name):
        super().__init__(name)

    def _value_function(self, batch_info):
        return batch_info[self.name]


class AveragingSupervisedMetric(BaseMetric):
    """ Base class for metrics that simply calculate the average over the epoch """
    def __init__(self, name):
        super().__init__(name)

        self.storage = []

    def calculate(self, batch_info):
        """ Calculate value of a metric """
        value = self._value_function(batch_info['data'], batch_info['target'], batch_info['output'])
        self.storage.append(value)

    def _value_function(self, x_input, y_true, y_pred):
        raise NotImplementedError

    def reset(self):
        """ Reset value of a metric """
        self.storage = []

    def value(self):
        """ Return current value for the metric """
        return np.mean(self.storage)



from .atari_wrappers import *
import numpy as np
import gym


class TimeLimit(gym.Wrapper):
    def __init__(self, env, max_episode_steps=None):
        super(TimeLimit, self).__init__(env)
        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = 0

    def step(self, ac):
        observation, reward, done, info = self.env.step(ac)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            done = True
            info['TimeLimit.truncated'] = True
        return observation, reward, done, info

    def reset(self, **kwargs):
        self._elapsed_steps = 0
        return self.env.reset(**kwargs)


class StochasticFrameSkip(gym.Wrapper):
    def __init__(self, env, n, stickprob):
        gym.Wrapper.__init__(self, env)
        self.n = n
        self.stickprob = stickprob
        self.curac = None
        self.rng = np.random.RandomState()
        self.supports_want_render = hasattr(env, "supports_want_render")

    def reset(self, **kwargs):
        self.curac = None
        return self.env.reset(**kwargs)

    def step(self, ac):
        done = False
        totrew = 0
        for i in range(self.n):
            # First step after reset, use action
            if self.curac is None:
                self.curac = ac
            # First substep, delay with probability=stickprob
            elif i==0:
                if self.rng.rand() > self.stickprob:
                    self.curac = ac
            # Second substep, new action definitely kicks in
            elif i==1:
                self.curac = ac
            if self.supports_want_render and i<self.n-1:
                ob, rew, done, info = self.env.step(self.curac, want_render=False)
            else:
                ob, rew, done, info = self.env.step(self.curac)
            totrew += rew
            if done: break
        return ob, totrew, done, info

    def seed(self, s):
        self.rng.seed(s)


class PartialFrameStack(gym.Wrapper):
    def __init__(self, env, k, channel=1):
        """
        Stack one channel (channel keyword) from previous frames
        """
        gym.Wrapper.__init__(self, env)
        shp = env.observation_space.shape
        self.channel = channel
        self.observation_space = gym.spaces.Box(low=0, high=255,
                                                shape=(shp[0], shp[1], shp[2] + k - 1),
                                                dtype=env.observation_space.dtype)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape

    def reset(self):
        ob = self.env.reset()
        assert ob.shape[2] > self.channel
        for _ in range(self.k):
            self.frames.append(ob)
        return self._get_ob()

    def step(self, ac):
        ob, reward, done, info = self.env.step(ac)
        self.frames.append(ob)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        return np.concatenate([frame if i==self.k-1 else frame[:,:,self.channel:self.channel+1]
                               for (i, frame) in enumerate(self.frames)], axis=2)


class Downsample(gym.ObservationWrapper):
    def __init__(self, env, ratio):
        """
        Downsample images by a factor of ratio
        """
        gym.ObservationWrapper.__init__(self, env)
        (oldh, oldw, oldc) = env.observation_space.shape
        newshape = (oldh//ratio, oldw//ratio, oldc)
        self.observation_space = spaces.Box(low=0, high=255,
                                            shape=newshape, dtype=np.uint8)

    def observation(self, frame):
        height, width, _ = self.observation_space.shape
        frame = cv2.resize(frame, (width, height), interpolation=cv2.INTER_AREA)
        if frame.ndim == 2:
            frame = frame[:,:,None]
        return frame


class Rgb2gray(gym.ObservationWrapper):
    def __init__(self, env):
        """
        Downsample images by a factor of ratio
        """
        gym.ObservationWrapper.__init__(self, env)
        (oldh, oldw, _oldc) = env.observation_space.shape
        self.observation_space = spaces.Box(low=0, high=255,
                                            shape=(oldh, oldw, 1), dtype=np.uint8)

    def observation(self, frame):
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        return frame[:,:,None]


class MovieRecord(gym.Wrapper):
    def __init__(self, env, savedir, k):
        gym.Wrapper.__init__(self, env)
        self.savedir = savedir
        self.k = k
        self.epcount = 0
    def reset(self):
        if self.epcount % self.k == 0:
            print('saving movie this episode', self.savedir)
            self.env.unwrapped.movie_path = self.savedir
        else:
            print('not saving this episode')
            self.env.unwrapped.movie_path = None
            self.env.unwrapped.movie = None
        self.epcount += 1
        return self.env.reset()


class AppendTimeout(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self.action_space = env.action_space
        self.timeout_space = gym.spaces.Box(low=np.array([0.0]), high=np.array([1.0]), dtype=np.float32)
        self.original_os = env.observation_space
        if isinstance(self.original_os, gym.spaces.Dict):
            import copy
            ordered_dict = copy.deepcopy(self.original_os.spaces)
            ordered_dict['value_estimation_timeout'] = self.timeout_space
            self.observation_space = gym.spaces.Dict(ordered_dict)
            self.dict_mode = True
        else:
            self.observation_space = gym.spaces.Dict({
                'original': self.original_os,
                'value_estimation_timeout': self.timeout_space
            })
            self.dict_mode = False
        self.ac_count = None
        while 1:
            if not hasattr(env, "_max_episode_steps"):  # Looking for TimeLimit wrapper that has this field
                env = env.env
                continue
            break
        self.timeout = env._max_episode_steps

    def step(self, ac):
        self.ac_count += 1
        ob, rew, done, info = self.env.step(ac)
        return self._process(ob), rew, done, info

    def reset(self):
        self.ac_count = 0
        return self._process(self.env.reset())

    def _process(self, ob):
        fracmissing = 1 - self.ac_count / self.timeout
        if self.dict_mode:
            ob['value_estimation_timeout'] = fracmissing
        else:
            return { 'original': ob, 'value_estimation_timeout': fracmissing }


class StartDoingRandomActionsWrapper(gym.Wrapper):
    """
    Warning: can eat info dicts, not good if you depend on them
    """
    def __init__(self, env, max_random_steps, on_startup=True, every_episode=False):
        gym.Wrapper.__init__(self, env)
        self.on_startup = on_startup
        self.every_episode = every_episode
        self.random_steps = max_random_steps
        self.last_obs = None
        if on_startup:
            self.some_random_steps()

    def some_random_steps(self):
        self.last_obs = self.env.reset()
        n = np.random.randint(self.random_steps)
        #print("running for random %i frames" % n)
        for _ in range(n):
            self.last_obs, _, done, _ = self.env.step(self.env.action_space.sample())
            if done: self.last_obs = self.env.reset()

    def reset(self):
        return self.last_obs

    def step(self, a):
        self.last_obs, rew, done, info = self.env.step(a)
        if done:
            self.last_obs = self.env.reset()
            if self.every_episode:
                self.some_random_steps()
        return self.last_obs, rew, done, info


def make_retro(*, game, state, max_episode_steps, **kwargs):
    import retro
    env = retro.make(game, state, **kwargs)
    env = StochasticFrameSkip(env, n=4, stickprob=0.25)
    if max_episode_steps is not None:
        env = TimeLimit(env, max_episode_steps=max_episode_steps)
    return env


def wrap_deepmind_retro(env, scale=True, frame_stack=4):
    """
    Configure environment for retro games, using config similar to DeepMind-style Atari in wrap_deepmind
    """
    env = WarpFrame(env)
    env = ClipRewardEnv(env)
    env = FrameStack(env, frame_stack)
    if scale:
        env = ScaledFloatFrame(env)
    return env


class SonicDiscretizer(gym.ActionWrapper):
    """
    Wrap a gym-retro environment and make it use discrete
    actions for the Sonic game.
    """
    def __init__(self, env):
        super(SonicDiscretizer, self).__init__(env)
        buttons = ["B", "A", "MODE", "START", "UP", "DOWN", "LEFT", "RIGHT", "C", "Y", "X", "Z"]
        actions = [['LEFT'], ['RIGHT'], ['LEFT', 'DOWN'], ['RIGHT', 'DOWN'], ['DOWN'],
                   ['DOWN', 'B'], ['B']]
        self._actions = []
        for action in actions:
            arr = np.array([False] * 12)
            for button in action:
                arr[buttons.index(button)] = True
            self._actions.append(arr)
        self.action_space = gym.spaces.Discrete(len(self._actions))

    def action(self, a): # pylint: disable=W0221
        return self._actions[a].copy()


class RewardScaler(gym.RewardWrapper):
    """
    Bring rewards to a reasonable scale for PPO.
    This is incredibly important and effects performance
    drastically.
    """
    def __init__(self, env, scale=0.01):
        super(RewardScaler, self).__init__(env)
        self.scale = scale

    def reward(self, reward):
        return reward * self.scale


class AllowBacktracking(gym.Wrapper):
    """
    Use deltas in max(X) as the reward, rather than deltas
    in X. This way, agents are not discouraged too heavily
    from exploring backwards if there is no way to advance
    head-on in the level.
    """
    def __init__(self, env):
        super(AllowBacktracking, self).__init__(env)
        self._cur_x = 0
        self._max_x = 0

    def reset(self, **kwargs): # pylint: disable=E0202
        self._cur_x = 0
        self._max_x = 0
        return self.env.reset(**kwargs)

    def step(self, action): # pylint: disable=E0202
        obs, rew, done, info = self.env.step(action)
        self._cur_x += rew
        rew = max(0, self._cur_x - self._max_x)
        self._max_x = max(self._max_x, self._cur_x)
        return obs, rew, done, info



import torch.utils.data as data


class DataFlow(data.Dataset):
    """ A dataset wrapping underlying data source with transformations """
    def __init__(self, dataset, transformations, tag):
        self.dataset = dataset

        if transformations is None:
            self.transformations = []
        else:
            self.transformations = [t for t in transformations if tag in t.tags]

        self.tag = tag

    def get_raw(self, index):
        return self.dataset[index]

    def __getitem__(self, index):
        raw_x, raw_y = self.dataset[index]

        for t in self.transformations:
            if t.mode == 'x':
                raw_x = t(raw_x)
            elif t.mode == 'y':
                raw_y = t(raw_y)
            elif t.mode == 'both':
                raw_x, raw_y = t(raw_x, raw_y)
            else:
                raise RuntimeError(f"Mode {t.mode} not recognized")

        return raw_x, raw_y

    def denormalize(self, datum, mode='x'):
        for t in self.transformations[::-1]:
            if t.mode == mode:
                datum = t.denormalize(datum)

        return datum

    def __len__(self):
        return len(self.dataset)



import numpy as np


class RunningMeanStd(object):
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    def __init__(self, epsilon=1e-4, shape=()):
        self.mean = np.zeros(shape, 'float64')
        self.var = np.ones(shape, 'float64')
        self.count = epsilon

    def update(self, x):
        batch_mean = np.mean(x, axis=0)
        batch_var = np.var(x, axis=0)
        batch_count = x.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(self, batch_mean, batch_var, batch_count):
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        M2 = m_a + m_b + np.square(delta) * self.count * batch_count / (self.count + batch_count)
        new_var = M2 / (self.count + batch_count)

        new_count = batch_count + self.count

        self.mean = new_mean
        self.var = new_var
        self.count = new_count


def test_runningmeanstd():
    for (x1, x2, x3) in [
        (np.random.randn(3), np.random.randn(4), np.random.randn(5)),
        (np.random.randn(3,2), np.random.randn(4,2), np.random.randn(5,2)),
    ]:

        rms = RunningMeanStd(epsilon=0.0, shape=x1.shape[1:])

        x = np.concatenate([x1, x2, x3], axis=0)
        ms1 = [x.mean(axis=0), x.var(axis=0)]
        rms.update(x1)
        rms.update(x2)
        rms.update(x3)
        ms2 = [rms.mean, rms.var]

        assert np.allclose(ms1, ms2)



import attr
import numpy as np
import sys
import torch
import tqdm

from vel.api import EpochInfo, BatchInfo
from vel.api.base import Model, ModelFactory
from vel.openai.baselines.common.vec_env import VecEnv
from vel.rl.api.base import ReinforcerBase, ReinforcerFactory, VecEnvFactory, ReplayEnvRollerBase, AlgoBase
from vel.rl.api.base.env_roller import ReplayEnvRollerFactory
from vel.rl.metrics import (
    FPSMetric, EpisodeLengthMetric, EpisodeRewardMetricQuantile,
    EpisodeRewardMetric, FramesMetric
)


@attr.s(auto_attribs=True)
class BufferedMixedPolicyIterationReinforcerSettings:
    """ Settings dataclass for a policy gradient reinforcer """
    discount_factor: float
    experience_replay: int = 1
    stochastic_experience_replay: bool = True


class BufferedMixedPolicyIterationReinforcer(ReinforcerBase):
    """
    A 'mixed' reinforcer that does both, on-policy learning from environment rollouts and off-policy learning
    from a replay buffer.

    Environments are rolled out in parallel and used as normally in off-policy learning.
    After that, each rollout is stored in a buffer that is sampled specified number of times per-environment
    for replay batches.
    """

    def __init__(self, device: torch.device, settings: BufferedMixedPolicyIterationReinforcerSettings, env: VecEnv,
                 model: Model, env_roller: ReplayEnvRollerBase, algo: AlgoBase) -> None:
        self.device = device
        self.settings = settings

        self.environment = env
        self._trained_model = model.to(self.device)

        self.env_roller = env_roller
        self.algo = algo

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        my_metrics = [
            FramesMetric("frames"),
            FPSMetric("fps"),
            EpisodeRewardMetric('PMM:episode_rewards'),
            EpisodeRewardMetricQuantile('P09:episode_rewards', quantile=0.9),
            EpisodeRewardMetricQuantile('P01:episode_rewards', quantile=0.1),
            EpisodeLengthMetric("episode_length")
        ]

        return my_metrics + self.algo.metrics() + self.env_roller.metrics()

    @property
    def model(self) -> Model:
        """ Model trained by this reinforcer """
        return self._trained_model

    def initialize_training(self, training_info):
        """ Prepare models for training """
        self.model.reset_weights()
        self.algo.initialize(self.settings, model=self.model, environment=self.environment, device=self.device)

    def train_epoch(self, epoch_info: EpochInfo, interactive=True):
        """ Train model on an epoch of a fixed number of batch updates """
        epoch_info.on_epoch_begin()

        if interactive:
            iterator = tqdm.trange(epoch_info.batches_per_epoch, file=sys.stdout, desc="Training", unit="batch")
        else:
            iterator = range(epoch_info.batches_per_epoch)

        for batch_idx in iterator:
            batch_info = BatchInfo(epoch_info, batch_idx)

            batch_info.on_batch_begin()
            self.train_batch(batch_info)
            batch_info.on_batch_end()

        epoch_info.result_accumulator.freeze_results()
        epoch_info.on_epoch_end()

    def train_batch(self, batch_info: BatchInfo):
        """ Single, most atomic 'step' of learning this reinforcer can perform """
        batch_info['sub_batch_data'] = []

        self.on_policy_train_batch(batch_info)

        if self.settings.experience_replay > 0 and self.env_roller.is_ready_for_sampling():
            if self.settings.stochastic_experience_replay:
                experience_replay_count = np.random.poisson(self.settings.experience_replay)
            else:
                experience_replay_count = self.settings.experience_replay

            for i in range(experience_replay_count):
                self.off_policy_train_batch(batch_info)

        # Even with all the experience replay, we count the single rollout as a single batch
        batch_info.aggregate_key('sub_batch_data')

    def on_policy_train_batch(self, batch_info: BatchInfo):
        """ Perform an 'on-policy' training step of evaluating an env and a single backpropagation step """
        self.model.eval()

        rollout = self.env_roller.rollout(batch_info, self.model)

        self.model.train()

        batch_result = self.algo.optimizer_step(
            batch_info=batch_info,
            device=self.device,
            model=self.model,
            rollout=rollout
        )

        batch_info['sub_batch_data'].append(batch_result)
        batch_info['frames'] = rollout.frames()
        batch_info['episode_infos'] = rollout.episode_information()

    def off_policy_train_batch(self, batch_info: BatchInfo):
        """ Perform an 'off-policy' training step of sampling the replay buffer and gradient descent """
        self.model.eval()

        rollout = self.env_roller.sample(batch_info, self.model)

        self.model.train()

        batch_result = self.algo.optimizer_step(
            batch_info=batch_info,
            device=self.device,
            model=self.model,
            rollout=rollout
        )

        batch_info['sub_batch_data'].append(batch_result)


class BufferedMixedPolicyIterationReinforcerFactory(ReinforcerFactory):
    """ Factory class for the PolicyGradientReplayBuffer factory """
    def __init__(self, settings, env_factory: VecEnvFactory, model_factory: ModelFactory,
                 env_roller_factory: ReplayEnvRollerFactory, algo: AlgoBase, parallel_envs: int, seed: int):
        self.settings = settings

        self.model_factory = model_factory
        self.env_factory = env_factory
        self.parallel_envs = parallel_envs
        self.env_roller_factory = env_roller_factory
        self.algo = algo
        self.seed = seed

    def instantiate(self, device: torch.device) -> ReinforcerBase:
        env = self.env_factory.instantiate(parallel_envs=self.parallel_envs, seed=self.seed)
        model = self.model_factory.instantiate(action_space=env.action_space)
        env_roller = self.env_roller_factory.instantiate(env, device, self.settings)

        return BufferedMixedPolicyIterationReinforcer(device, self.settings, env, model, env_roller, self.algo)


def create(model_config, model, vec_env, algo, env_roller,
           parallel_envs, discount_factor,
           experience_replay=1, stochastic_experience_replay=True):
    """ Create a policy gradient reinforcer - factory """
    settings = BufferedMixedPolicyIterationReinforcerSettings(
        discount_factor=discount_factor,
        experience_replay=experience_replay,
        stochastic_experience_replay=stochastic_experience_replay
    )

    return BufferedMixedPolicyIterationReinforcerFactory(
        settings,
        env_factory=vec_env,
        model_factory=model,
        parallel_envs=parallel_envs,
        env_roller_factory=env_roller,
        algo=algo,
        seed=model_config.seed
    )



import numpy as np
import torch

from vel.api.base import Schedule
from vel.api.metrics import AveragingNamedMetric
from vel.rl.api import Rollout, Transitions
from vel.rl.api.base import ReplayEnvRollerBase, ReplayEnvRollerFactory
from vel.rl.buffers.deque_backend import DequeBufferBackend


class DequeReplayRollerEpsGreedy(ReplayEnvRollerBase):
    """
    Environment roller for action-value models using experience replay.
    Simplest buffer implementation just holding up to given number of samples.

    Because framestack is implemented directly in the buffer, we can use *much* less space to hold samples in
    memory for very little additional cost.
    """

    def __init__(self, environment, device, epsilon_schedule: Schedule, batch_size: int,
                 buffer_capacity: int, buffer_initial_size: int, frame_stack: int):
        self.epsilon_schedule = epsilon_schedule
        self.batch_size = batch_size
        self.buffer_capacity = buffer_capacity
        self.buffer_initial_size = buffer_initial_size
        self.frame_stack = frame_stack

        self.device = device
        self._environment = environment
        self.backend = DequeBufferBackend(
            buffer_capacity=self.buffer_capacity,
            observation_space=environment.observation_space,
            action_space=environment.action_space
        )

        self.last_observation = self.environment.reset()

    @property
    def environment(self):
        """ Return environment of this env roller """
        return self._environment

    def is_ready_for_sampling(self) -> bool:
        """ If buffer is ready for drawing samples from it (usually checks if there is enough data) """
        return self.backend.current_size >= self.buffer_initial_size

    def epsgreedy_action(self, policy_samples, epsilon):
        """ Sample e-greedy action using curreny policy and epsilon value """
        random_samples = torch.randint_like(policy_samples, self.environment.action_space.n)
        selector = torch.rand_like(random_samples, dtype=torch.float32)
        return torch.where(selector > epsilon, policy_samples, random_samples)

    @torch.no_grad()
    def rollout(self, batch_info, model) -> Rollout:
        """ Roll-out the environment and return it """
        epsilon_value = self.epsilon_schedule.value(batch_info['progress'])
        batch_info['epsilon'] = epsilon_value

        last_observation = np.concatenate([
            self.backend.get_frame(self.backend.current_idx, self.frame_stack - 1),
            self.last_observation
        ], axis=-1)

        observation_tensor = torch.from_numpy(last_observation[None]).to(self.device)
        step = model.step(observation_tensor)

        epsgreedy_step = self.epsgreedy_action(step['actions'], epsilon_value)
        action = epsgreedy_step.item()

        observation, reward, done, info = self.environment.step(action)
        self.backend.store_transition(self.last_observation, action, reward, done)

        # Usual, reset on done
        if done:
            observation = self.environment.reset()

        self.last_observation = observation

        return Transitions(
            size=1,
            environment_information=[info],
            transition_tensors={
                'actions': epsgreedy_step.unsqueeze(0),
                'values': step['values']
            },
            extra_data={
                'epsilon': epsilon_value
            }
        )

    def metrics(self):
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("epsilon"),
        ]

    def sample(self, batch_info, model) -> Transitions:
        """ Sample experience from replay buffer and return a batch """
        indexes = self.backend.sample_batch_uniform(self.batch_size, self.frame_stack)
        batch = self.backend.get_batch(indexes, self.frame_stack)

        observations = torch.from_numpy(batch['states']).to(self.device)
        observations_plus1 = torch.from_numpy(batch['states+1']).to(self.device)
        dones = torch.from_numpy(batch['dones'].astype(np.float32)).to(self.device)
        rewards = torch.from_numpy(batch['rewards'].astype(np.float32)).to(self.device)
        actions = torch.from_numpy(batch['actions']).to(self.device)

        return Transitions(
            size=self.batch_size,
            environment_information=None,
            transition_tensors={
                'observations': observations,
                'observations_next': observations_plus1,
                'dones': dones,
                'rewards': rewards,
                'actions': actions,
                'weights': torch.ones_like(rewards)
            }
        )


class DequeReplayRollerEpsGreedyFactory(ReplayEnvRollerFactory):
    """ Factory class for DequeReplayQRoller """
    def __init__(self, epsilon_schedule: Schedule, buffer_capacity: int, buffer_initial_size: int,
                 frame_stack: int=1):
        self.buffer_capacity = buffer_capacity
        self.epsilon_schedule = epsilon_schedule
        self.buffer_initial_size = buffer_initial_size
        self.frame_stack = frame_stack

    def instantiate(self, environment, device, settings) -> ReplayEnvRollerBase:
        return DequeReplayRollerEpsGreedy(
            environment, device, self.epsilon_schedule, settings.batch_size,
            self.buffer_capacity, self.buffer_initial_size, self.frame_stack
        )


def create(epsilon_schedule: Schedule, buffer_capacity: int, buffer_initial_size: int, frame_stack: int=1):
    return DequeReplayRollerEpsGreedyFactory(
        epsilon_schedule=epsilon_schedule,
        buffer_capacity=buffer_capacity,
        buffer_initial_size=buffer_initial_size,
        frame_stack=frame_stack
    )



import vel.util.intepolate as interpolate

from vel.api.base import Schedule


class LinearAndConstantSchedule(Schedule):
    """
    Interpolate variable linearly between start value and final values
    only up to given point and return constant value afterwards.
    """

    def __init__(self, initial_value, final_value, end_of_interpolation):
        self.initial_value = initial_value
        self.final_value = final_value
        self.end_of_interpolation = end_of_interpolation

    def value(self, progress_indicator):
        """ Interpolate linearly between start and end """
        if progress_indicator <= self.end_of_interpolation:
            return interpolate.interpolate_linear_single(
                self.initial_value, self.final_value, progress_indicator/self.end_of_interpolation
            )
        else:
            return self.final_value


def create(initial_value, final_value, end_of_interpolation):
    """ Vel creation function """
    return LinearAndConstantSchedule(initial_value, final_value, end_of_interpolation)



import numpy as np

import vel.api.data as data


class ToArray(data.Augmentation):
    """ Scales the image so that the smallest axis is of 'size'. """
    def __init__(self, mode='x', tags=None):
        super().__init__(mode, tags)

    def __call__(self, x_data):
        array = np.array(x_data)

        if array.dtype == np.uint8:
            return array.astype(np.float32) / 255.0
        else:
            return array


def create(mode='x', tags=None):
    return ToArray(mode, tags)



from vel.openai.baselines.common.vec_env import VecEnv
from vel.openai.baselines.common.atari_wrappers import FrameStack
from vel.openai.baselines.common.vec_env.subproc_vec_env import SubprocVecEnv
from vel.openai.baselines.common.vec_env.vec_normalize import VecNormalize
from vel.openai.baselines.common.vec_env.vec_frame_stack import VecFrameStack

from vel.rl.api.base import VecEnvFactory


class SubprocVecEnvWrapper(VecEnvFactory):
    """ Wrapper for an environment to create sub-process vector environment """

    def __init__(self, env, frame_history=None, normalize=False):
        self.env = env
        self.frame_history = frame_history
        self.normalize = normalize

    def instantiate(self, parallel_envs, seed=0, preset='default') -> VecEnv:
        """ Make parallel environments """
        envs = SubprocVecEnv([self._creation_function(i, seed, preset) for i in range(parallel_envs)])

        if self.normalize:
            envs = VecNormalize(envs)

        if self.frame_history is not None:
            envs = VecFrameStack(envs, self.frame_history)

        return envs

    def instantiate_single(self, seed=0, preset='default'):
        """ Create a new VecEnv instance - single """
        env = self.env.instantiate(seed=seed, serial_id=0, preset=preset)

        if self.normalize:
            raise NotImplementedError

        if self.frame_history is not None:
            env = FrameStack(env, self.frame_history)

        return env

    def _creation_function(self, idx, seed, preset):
        """ Helper function to create a proper closure around supplied values """
        return lambda: self.env.instantiate(seed=seed, serial_id=idx, preset=preset)


def create(env, frame_history=None, normalize=False):
    return SubprocVecEnvWrapper(env, frame_history=frame_history, normalize=normalize)



import sys
import torch
import tqdm
import typing

from .info import BatchInfo, EpochInfo, TrainingInfo


class Learner:
    """ Manages training process of a single model """
    def __init__(self, device: torch.device, model, max_grad_norm: typing.Optional[float]=None):
        self.device = device
        self.model = model.to(device)
        self.max_grad_norm = max_grad_norm

    def metrics(self):
        """ Return metrics for given learner/model """
        return self.model.metrics()

    def summary(self):
        """ Print summary for given learner/model """
        return self.model.summary()

    def train(self):
        """ Set model in the training mode """
        return self.model.train()

    def eval(self):
        """ Set model in the evaluation mode """
        return self.model.eval()

    def number_of_parameters(self):
        """ Count model parameters """
        return sum(p.numel() for p in self.model.parameters())

    def initialize_training(self, training_info: TrainingInfo):
        """ Prepare for training """
        self.model.reset_weights()

    def run_epoch(self, epoch_info: EpochInfo, source: 'vel.api.base.Source'):
        """ Run full epoch of learning """
        epoch_info.on_epoch_begin()

        lr = epoch_info.optimizer.param_groups[-1]['lr']
        print("|-------- Epoch {:06} Lr={:.6f} ----------|".format(epoch_info.global_epoch_idx, lr))

        self.train_epoch(epoch_info, source)
        epoch_info.result_accumulator.freeze_results('train')

        self.validation_epoch(epoch_info, source)
        epoch_info.result_accumulator.freeze_results('val')

        epoch_info.on_epoch_end()

    def train_epoch(self, epoch_info, source: 'vel.api.base.Source', interactive=True):
        """ Run a single training epoch """
        self.train()

        if interactive:
            iterator = tqdm.tqdm(source.train_loader(), desc="Training", unit="iter", file=sys.stdout)
        else:
            iterator = source.train_loader()

        for batch_idx, (data, target) in enumerate(iterator):
            batch_info = BatchInfo(epoch_info, batch_idx)

            batch_info.on_batch_begin()
            self.train_batch(batch_info, data, target)
            batch_info.on_batch_end()

            iterator.set_postfix(loss=epoch_info.result_accumulator.intermediate_value('loss'))

    def validation_epoch(self, epoch_info, source: 'vel.api.base.Source'):
        """ Run a single evaluation epoch """
        self.eval()

        iterator = tqdm.tqdm(source.val_loader(), desc="Validation", unit="iter", file=sys.stdout)

        with torch.no_grad():
            for batch_idx, (data, target) in enumerate(iterator):
                batch_info = BatchInfo(epoch_info, batch_idx)

                batch_info.on_validation_batch_begin()
                self.feed_batch(batch_info, data, target)
                batch_info.on_validation_batch_end()

    def feed_batch(self, batch_info, data, target):
        """ Run single batch of data """
        data, target = data.to(self.device), target.to(self.device)
        output, loss = self.model.loss(data, target)

        # Store extra batch information for calculation of the statistics
        batch_info['data'] = data
        batch_info['target'] = target
        batch_info['output'] = output
        batch_info['loss'] = loss

        return loss

    def train_batch(self, batch_info, data, target):
        """ Train single batch of data """
        batch_info.optimizer.zero_grad()
        loss = self.feed_batch(batch_info, data, target)
        loss.backward()

        if self.max_grad_norm is not None:
            batch_info['grad_norm'] = torch.nn.utils.clip_grad_norm_(
                filter(lambda p: p.requires_grad, self.model.parameters()),
                max_norm=self.max_grad_norm
            )

        batch_info.optimizer.step()



#!/usr/bin/env python
import argparse
import datetime as dtm

from vel.api import ModelConfig
from vel.util.random import set_seed
from vel.internals.parser import Parser


def main():
    """ Paperboy entry point - parse the arguments and run a command """
    parser = argparse.ArgumentParser(description='Paperboy deep learning launcher')

    parser.add_argument('config', metavar='FILENAME', help='Configuration file for the run')
    parser.add_argument('command', metavar='COMMAND', help='A command to run')
    parser.add_argument('varargs', nargs='*', metavar='VARARGS', help='Extra options to the command')
    parser.add_argument('-r', '--run_number', type=int, default=0, help="A run number")
    parser.add_argument('-d', '--device', default='cuda', help="A device to run the model on")
    parser.add_argument('-s', '--seed', type=int, default=None, help="Random seed for the project")
    parser.add_argument(
        '-p', '--param', type=str, metavar='NAME=VALUE', action='append', default=[],
        help="Configuration parameters"
    )
    parser.add_argument(
        '--continue', action='store_true', default=False, help="Continue previously started learning process"
    )

    args = parser.parse_args()

    model_config = ModelConfig.from_file(
        args.config, args.run_number, continue_training=getattr(args, 'continue'), device=args.device, seed=args.seed,
        params={k: v for (k, v) in (Parser.parse_equality(eq) for eq in args.param)}
    )

    # Set seed already in the launcher
    set_seed(model_config.seed)

    model_config.banner(args.command)
    model_config.run_command(args.command, args.varargs)
    model_config.quit_banner()


if __name__ == '__main__':
    main()



from torchvision import datasets, transforms


from vel.api.base import TrainingData


def create(batch_size, model_config, normalize=True, num_workers=0):
    """ Create a MNIST dataset, normalized """
    path = model_config.data_dir('mnist')

    train_dataset = datasets.MNIST(path, train=True, download=True)
    test_dataset = datasets.MNIST(path, train=False, download=True)

    if normalize:
        train_data = train_dataset.train_data
        mean_value = (train_data.double() / 255).mean().item()
        std_value = (train_data.double() / 255).std().item()

        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((mean_value,), (std_value,))
        ])
    else:
        transform = transforms.ToTensor()

    train_dataset.transform = transform
    test_dataset.transform = transform

    return TrainingData(
        train_dataset,
        test_dataset,
        num_workers=num_workers,
        batch_size=batch_size
    )



import pandas as pd


class DummyBackend:
    """ Storage backend to store all experiment data in /dev/null """

    def __init__(self):
        pass

    def clean(self, initial_epoch):
        """ Remove entries from database that would get overwritten """
        pass

    def get_frame(self):
        """ Get a dataframe of metrics from this storage """
        return pd.DataFrame()

    def store(self, metrics):
        """ Store metrics in a datastore """
        pass

    def store_config(self, configuration):
        """ Store model parameters """
        pass


def create():
    """ Create a mongodb storage object """
    return DummyBackend()



import gym

from vel.api.base import LinearBackboneModel, Model, ModelFactory
from vel.rl.api import Evaluator, Rollout
from vel.rl.modules.q_head import QHead


class QModelEvaluator(Evaluator):
    """ Evaluate simple q-model """
    def __init__(self, model: 'QModel', rollout: Rollout):
        super().__init__(rollout)
        self.model = model

    @Evaluator.provides('model:q')
    def model_q(self):
        """ Action values for all (discrete) actions """
        observations = self.get('rollout:observations')
        return self.model(observations)

    @Evaluator.provides('model:action:q')
    def model_action_q(self):
        """ Action values for all (discrete) actions """
        q = self.get('model:q')
        actions = self.get('rollout:actions')
        return q.gather(1, actions.unsqueeze(1)).squeeze(1)

    @Evaluator.provides('model:q_next')
    def model_q_next(self):
        """ Action values for all (discrete) actions """
        observations = self.get('rollout:observations_next')
        return self.model(observations)


class QModel(Model):
    """ Wraps a backbone model into API we need for Deep Q-Learning """

    def __init__(self, backbone: LinearBackboneModel, action_space: gym.Space):
        super().__init__()

        self.backbone = backbone
        self.q_head = QHead(input_dim=backbone.output_dim, action_space=action_space)

    def forward(self, observations):
        """ Model forward pass """
        base_output = self.backbone(observations)
        q_values = self.q_head(base_output)
        return q_values

    def reset_weights(self):
        """ Initialize weights to reasonable defaults """
        self.backbone.reset_weights()
        self.q_head.reset_weights()

    def step(self, observations):
        """ Sample action from an action space for given state """
        q_values = self(observations)

        return {
            'actions': self.q_head.sample(q_values),
            'values': q_values
        }

    def evaluate(self, rollout: Rollout) -> Evaluator:
        """ Evaluate model on a rollout """
        return QModelEvaluator(self, rollout)


class QModelFactory(ModelFactory):
    """ Factory class for q-learning models """
    def __init__(self, backbone: ModelFactory):
        self.backbone = backbone

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        backbone = self.backbone.instantiate(**extra_args)
        return QModel(backbone, extra_args['action_space'])


def create(backbone: ModelFactory):
    """ Q-Learning model factory """
    return QModelFactory(backbone=backbone)



import torch
import typing
import torch.autograd
import torch.nn.functional as F

from vel.rl.api.base import OptimizerAlgoBase
from vel.api.metrics.averaging_metric import AveragingNamedMetric


class DeepDeterministicPolicyGradient(OptimizerAlgoBase):
    """ Deep Deterministic Policy Gradient (DDPG) - policy gradient calculations """

    def __init__(self, model_factory, tau: float, max_grad_norm: typing.Optional[float]=None):
        super().__init__(max_grad_norm)

        self.model_factory = model_factory
        self.tau = tau

        self.discount_factor = None
        self.target_model = None

    def initialize(self, settings, model, environment, device):
        """ Initialize algo from reinforcer settings """
        self.discount_factor = settings.discount_factor

        self.target_model = self.model_factory.instantiate(action_space=environment.action_space).to(device)
        self.target_model.load_state_dict(model.state_dict())

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        rollout = rollout.to_transitions()

        dones = rollout.batch_tensor('dones')
        rewards = rollout.batch_tensor('rewards')
        observations_next = rollout.batch_tensor('observations_next')
        actions = rollout.batch_tensor('actions')
        observations = rollout.batch_tensor('observations')

        # Calculate value loss - or critic loss
        with torch.no_grad():
            target_next_value = self.target_model.value(observations_next)
            target_value = rewards + (1.0 - dones) * self.discount_factor * target_next_value

        # Value estimation error vs the target network
        model_value = model.value(observations, actions)
        value_loss = F.mse_loss(model_value, target_value)

        # It may seem a bit tricky what I'm doing here, but the underlying idea is simple
        # All other implementations I found keep two separate optimizers for actor and critic
        # and update them separately
        # What I'm trying to do is to optimize them both with a single optimizer
        # but I need to make sure gradients flow correctly
        # From critic loss to critic network only and from actor loss to actor network only

        # Backpropagate value loss to critic only
        value_loss.backward()

        model_action = model.action(observations)
        model_action_value = model.value(observations, model_action)

        policy_loss = -model_action_value.mean()

        model_action_grad = torch.autograd.grad(policy_loss, model_action)[0]

        # Backpropagate actor loss to actor only
        model_action.backward(gradient=model_action_grad)

        return {
            'policy_loss': policy_loss.item(),
            'value_loss': value_loss.item(),
        }

    def post_optimization_step(self, batch_info, device, model, rollout):
        """ Steps to take after optimization has been done"""
        # Update target model
        for model_param, target_param in zip(model.parameters(), self.target_model.parameters()):
            # EWMA average model update
            target_param.data.mul_(1 - self.tau).add_(model_param.data * self.tau)

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("value_loss"),
            AveragingNamedMetric("policy_loss"),
        ]


def create(model, tau: float, max_grad_norm: float=None):
    return DeepDeterministicPolicyGradient(
        tau=tau,
        model_factory=model,
        max_grad_norm=max_grad_norm
    )



import typing


class CheckpointStrategy:
    """ Base class for various checkpoint strategies """
    def should_delete_previous_checkpoint(self, epoch_idx) -> bool:
        """ Should previous checkpoint be deleted or stored """
        return True

    def should_store_best_checkpoint(self, epoch_idx, metrics) -> bool:
        """ Should we store current checkpoint as the best """
        return False

    def store_best_checkpoint_idx(self, epoch_idx) -> None:
        """ Should we store current checkpoint as the best """
        pass

    @property
    def current_best_checkpoint_idx(self) -> typing.Union[int, None]:
        return None

    def write_state_dict(self, hidden_state_dict): pass

    def restore(self, hidden_state_dict): pass





import torch.utils.data as data


class Source:
    """ Source of data for supervised learning algorithms """
    def __init__(self):
        pass

    def train_loader(self):
        """ PyTorch loader of training data """
        raise NotImplementedError

    def val_loader(self):
        """ PyTorch loader of validation data """
        raise NotImplementedError

    def train_dataset(self):
        """ Return the training dataset """
        raise NotImplementedError

    def val_dataset(self):
        """ Return the validation dataset """
        raise NotImplementedError

    def train_iterations_per_epoch(self):
        """ Return number of iterations per epoch """
        raise NotImplementedError

    def val_iterations_per_epoch(self):
        """ Return number of iterations per epoch - validation """
        raise NotImplementedError


class TrainingData(Source):
    """ Most common source of data combining a basic datasource and sampler """
    def __init__(self, train_source, val_source, num_workers, batch_size, augmentations=None):
        import vel.api.data as vel_data

        super().__init__()

        self.train_source = train_source
        self.val_source = val_source

        self.num_workers = num_workers
        self.batch_size = batch_size

        self.augmentations = augmentations

        # Derived values
        self.train_ds = vel_data.DataFlow(self.train_source, augmentations, tag='train')
        self.val_ds = vel_data.DataFlow(self.val_source, augmentations, tag='val')

        self._train_loader = data.DataLoader(
            self.train_ds, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )

        self._val_loader = data.DataLoader(
            self.val_ds, batch_size=batch_size, shuffle=False, num_workers=num_workers
        )

    def train_loader(self):
        """ PyTorch loader of training data """
        return self._train_loader

    def val_loader(self):
        """ PyTorch loader of validation data """
        return self._val_loader

    def train_dataset(self):
        """ Return the training dataset """
        return self.train_ds

    def val_dataset(self):
        """ Return the validation dataset """
        return self.val_ds

    def train_iterations_per_epoch(self):
        """ Return number of iterations per epoch """
        return len(self._train_loader)

    def val_iterations_per_epoch(self):
        """ Return number of iterations per epoch - validation """
        return len(self._val_loader)



import vel.api.base as base

from vel.api import BatchInfo, TrainingInfo


class LinearBatchScaler(base.Callback):
    """ Scales linearly LR from maximum value to 0 through every batch of training """
    def __init__(self, optimizer):
        self.optimizer = optimizer
        self.starting_lr = None

    def on_initialization(self, training_info: TrainingInfo):
        self.starting_lr = [p['lr'] for p in self.optimizer.param_groups]

    def write_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        hidden_state_dict['linear_batch_scaler/starting_lr'] = self.starting_lr

    def load_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        self.starting_lr = hidden_state_dict['linear_batch_scaler/starting_lr']

    def on_batch_begin(self, batch_info: BatchInfo):
        for starting_lr, param_group in zip(self.starting_lr, self.optimizer.param_groups):
            param_group['lr'] = starting_lr * (1.0 - batch_info['progress'])


class LinearBatchScalerFactory(base.SchedulerFactory):
    """ Factory class for linear batch scaler scheduler """
    def instantiate(self, optimizer, last_epoch=-1) -> LinearBatchScaler:
        return LinearBatchScaler(optimizer)


def create():
    return LinearBatchScalerFactory()





from setuptools import setup, find_packages

long_description = """
This repository is my project to bring velocity to deep-learning research, by providing tried and tested large pool of
prebuilt components that are known to be working well together.

I would like to minimize time to market of new projects, ease experimentation and provide bits of experiment management
to bring some order to the data science workflow.

Ideally, for most applications it should be enough to write a config file wiring existing components together.
If that's not the case writing bits of custom code shouldn't be unnecessarily complex.

This repository is still in an early stage of that journey but it will grow as I'll be putting some work into it.
"""


setup(
    name='vel',
    version='0.2.1',
    description="Velocity in deep-learning research",
    long_description=long_description,
    url='https://github.com/MillionIntegrals/vel',
    author='Jerry Tworek',
    author_email='jerry@millionintegrals.com',
    license='MIT',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    python_requires='>=3.6',
    install_requires=[
        'attrs',
        'pyyaml',
        'numpy',
        'pandas',
        'scikit-learn',
        'torch == 0.4.1',
        'torchvision',
        'opencv-python',
        'tqdm'
    ],
    extras_require={
        'visdom': ['visdom'],
        'mongo': ['pymongo', 'dnspython'],
        'gym': ['gym[atari,box2d,classic_control]'],
        'mujoco': ['gym[mujoco,robotics]'],
        'dev': ['nose']
    },
    entry_points={
        'console_scripts': [
            'vel = vel.launcher:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    scripts=[]
)



import vel.api.base as base

from vel.api import EpochInfo


class StdoutStreaming(base.Callback):
    """ Stream results to stdout """
    def on_epoch_end(self, epoch_info: EpochInfo):
        if epoch_info.training_info.run_name:
            print(f"=>>>>>>>>>> EPOCH {epoch_info.global_epoch_idx} [{epoch_info.training_info.run_name}]")
        else:
            print(f"=>>>>>>>>>> EPOCH {epoch_info.global_epoch_idx}")

        if any(':' not in x for x in epoch_info.result.keys()):
            self._print_metrics_line(epoch_info.result, head=None)

        head_set = sorted({x.split(':')[0] + ':' for x in epoch_info.result.keys() if ':' in x})

        for head in head_set:
            if any(x.startswith(head) for x in epoch_info.result.keys()):
                self._print_metrics_line(epoch_info.result, head)

        print(f"=>>>>>>>>>> DONE")

    @staticmethod
    def _print_metrics_line(metrics, head=None):
        if head is None:
            head = 'Metrics:'

            metrics_list = [
                "{} {:.06f}".format(k, metrics[k])
                for k in sorted([k for k in metrics.keys() if ':' not in k])
            ]
        else:
            metrics_list = [
                "{} {:.06f}".format(k.split(':')[1], metrics[k])
                for k in sorted([k for k in metrics.keys() if k.startswith(head)])
            ]

        print('{0: <10}'.format(head.capitalize()), " ".join(metrics_list))


def create():
    """ Vel create function """
    return StdoutStreaming()



import torch


def one_hot_encoding(input_tensor, num_labels):
    """ One-hot encode labels from input """
    return torch.eye(num_labels, device=input_tensor.device)[input_tensor.to(torch.long)]


def merge_first_two_dims(tensor):
    """ Reshape tensor to merge first two dimensions """
    shape = tensor.shape
    batch_size = shape[0] * shape[1]
    new_shape = tuple([batch_size] + list(shape[2:]))
    return tensor.view(new_shape)



import torch
import torch.nn as nn
import torch.nn.init as init

import gym.spaces as spaces


class QHead(nn.Module):
    """ Network head calculating Q-function value for each (discrete) action. """
    def __init__(self, input_dim, action_space):
        super().__init__()

        # Q-function requires a discrete action space
        assert isinstance(action_space, spaces.Discrete)

        self.linear_layer = nn.Linear(input_dim, action_space.n)
        self.action_space = action_space

    def reset_weights(self):
        init.orthogonal_(self.linear_layer.weight, gain=1.0)
        init.constant_(self.linear_layer.bias, 0.0)

    def forward(self, input_data):
        return self.linear_layer(input_data)

    def sample(self, q_values):
        """ Sample from epsilon-greedy strategy with given q-values """
        return q_values.argmax(dim=1)




import cv2
import math
import numpy as np


def crop_square(im, r, c, sz):
    '''
    crop image into a square of size sz,
    '''
    return im[r:r+sz, c:c+sz]


def crop(im, r, c, sz_h, sz_w):
    '''
    crop image into a square of size sz,
    '''
    return im[r:r+sz_h, c:c+sz_w]


def center_crop(im, min_sz=None):
    """ Returns a center crop of an image"""
    # return F.center_crop(im, min_sz)
    r,c,*_ = im.shape
    if min_sz is None: min_sz = min(r,c)
    start_r = math.ceil((r-min_sz)/2)
    start_c = math.ceil((c-min_sz)/2)
    return crop_square(im, start_r, start_c, min_sz)


def scale_to(x, ratio, targ):
    '''Calculate dimension of an image during scaling with aspect ratio'''
    return max(math.floor(x*ratio), targ)


def scale_min(im, targ, interpolation=cv2.INTER_AREA):
    """ Scales the image so that the smallest axis is of size targ.

    Arguments:
        im (array): image
        targ (int): target size
    """
    r,c,*_ = im.shape

    ratio = targ/min(r,c)

    sz = (scale_to(c, ratio, targ), scale_to(r, ratio, targ))

    # return im.resize(sz, resample=interpolation)
    return cv2.resize(im, sz, interpolation=interpolation)


# def rotate_img(img, deg, interpolation=Image.BICUBIC):
#     """ Rotate image by given angle """
#     return img.rotate(deg, resample=interpolation)

def rotate_img(im, deg, mode=cv2.BORDER_CONSTANT, interpolation=cv2.INTER_AREA):
    """ Rotates an image by deg degrees

    Arguments:
        deg (float): degree to rotate.
    """
    r,c,*_ = im.shape
    M = cv2.getRotationMatrix2D((c//2,r//2),deg,1)
    return cv2.warpAffine(im,M,(c,r), borderMode=mode, flags=cv2.WARP_FILL_OUTLIERS+interpolation)


def pad(img, pad, mode=cv2.BORDER_REFLECT):
    return cv2.copyMakeBorder(img, pad, pad, pad, pad, mode)


def mode_to_cv2(mode='constant'):
    if mode == 'constant':
        return cv2.BORDER_CONSTANT
    if mode == 'reflect':
        return cv2.BORDER_REFLECT

    raise ValueError(f"Invalid mode {mode}")


def lighting(im, b, c):
    ''' adjusts image's balance and contrast'''
    if b==0 and c==1: return im
    mu = np.average(im)
    return np.clip((im-mu)*c+mu+b,0.,1.).astype(np.float32)



from vel.api.base import Source


class ModelSummary:
    """ Just print model summary """
    def __init__(self, model, source: Source):
        self.model = model
        self.source = source

    def run(self, *args):
        """ Print model summary """
        if self.source is None:
            self.model.summary()
        else:
            x_data, y_data = next(iter(self.source.train_loader()))
            self.model.summary(input_size=x_data.shape[1:])


def create(model, source=None):
    """ Vel creation function """
    return ModelSummary(model, source)



import numpy as np


def better(old_value, new_value, mode):
    """ Check if new value is better than the old value"""
    if (old_value is None or np.isnan(old_value)) and (new_value is not None and not np.isnan(new_value)):
        return True

    if mode == 'min':
        return new_value < old_value
    elif mode == 'max':
        return new_value > old_value
    else:
        raise RuntimeError(f"Mode '{mode}' value is not supported")


"""
Slightly modified version of:
https://github.com/pytorch/vision/blob/master/torchvision/transforms/transforms.py
"""

import numbers
import random

import vel.api.data as data


class RandomCrop(data.Augmentation):
    """Crop the given PIL Image at a random location.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is 0, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively.
        pad_if_needed (boolean): It will pad the image if smaller than the
            desired size to avoid raising an exception.
    """

    def __init__(self, size, padding=0, padding_mode='constant', pad_if_needed=False, mode='x', tags=None):
        super().__init__(mode, tags)

        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

        self.padding = padding
        self.padding_mode = padding_mode
        self.padding_mode_cv = data.mode_to_cv2(self.padding_mode)
        self.pad_if_needed = pad_if_needed

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h, *_ = img.shape

        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped.

        Returns:
            PIL Image: Cropped image.
        """
        if self.padding > 0:
            img = data.pad(img, self.padding, mode=self.padding_mode_cv)

        # pad the width if needed
        if self.pad_if_needed and img.size[0] < self.size[1]:
            img = data.pad(img, (int((1 + self.size[1] - img.size[0]) / 2), 0), mode=self.padding_mode_cv)

        # pad the height if needed
        if self.pad_if_needed and img.size[1] < self.size[0]:
            img = data.pad(img, (0, int((1 + self.size[0] - img.size[1]) / 2)), mode=self.padding_mode_cv)

        i, j, h, w = self.get_params(img, self.size)

        return data.crop(img, j, i, w, h)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)


def create(width, height, padding=0, padding_mode='constant', mode='x', tags=None):
    return RandomCrop(size=(width, height), padding=padding, padding_mode=padding_mode, mode=mode, tags=tags)



import vel.api.base as base


class FreezePhase(base.EmptyTrainPhase):
    """ Freeze the model """

    def set_up_phase(self, training_info, model, source):
        """ Freeze the model """
        model.freeze()


def create():
    """ Vel creation function """
    return FreezePhase()



import numpy as np

import vel.api.base as base
import vel.util.intepolate as interp

from vel.api import BatchInfo, EpochInfo, TrainingInfo


class CycleCallback(base.Callback):
    """ A callback that manages setting the proper learning rate """

    def __init__(self, optimizer, max_lr, min_lr, cycles, cycle_len=1, cycle_mult=1, init_iter=0, init_lr=0,
                 interpolate='linear'):
        self.max_lr = max_lr
        self.min_lr = min_lr

        self.cycles = cycles
        self.cycle_len = cycle_len
        self.cycle_mult = cycle_mult

        self.init_iter = init_iter
        self.init_lr = init_lr

        if cycle_mult > 1:
            self.epochs = self.cycle_len * (self.cycle_mult ** self.cycles - 1) // (self.cycle_mult - 1)
        else:
            self.epochs = self.cycle_mult * self.cycles

        self.optimizer = optimizer
        self.interpolate = interpolate

        # self.current_cycle = None
        self.cycle_dict, self.cycle_lengths, self.cycle_starts = self._init_cycle_dict()

    def _init_cycle_dict(self):
        """ Populate a cycle dict """
        dict_arr = np.zeros(self.epochs, dtype=int)
        length_arr = np.zeros(self.epochs, dtype=int)
        start_arr = np.zeros(self.epochs, dtype=int)

        c_len = self.cycle_len
        idx = 0

        for i in range(self.cycles):
            current_start = idx
            for j in range(c_len):
                dict_arr[idx] = i
                length_arr[idx] = c_len
                start_arr[idx] = current_start
                idx += 1

            c_len *= self.cycle_mult

        return dict_arr, length_arr, start_arr

    def on_batch_begin(self, batch_info: BatchInfo):
        """ Set proper learning rate """
        cycle_length = self.cycle_lengths[batch_info.local_epoch_number - 1]
        cycle_start = self.cycle_starts[batch_info.local_epoch_number - 1]

        numerator = (batch_info.local_epoch_number - cycle_start - 1) * batch_info.batches_per_epoch + batch_info.batch_number
        denominator = cycle_length * batch_info.batches_per_epoch

        interpolation_number = numerator / denominator

        if cycle_start == 0 and numerator < self.init_iter:
            lr = self.init_lr
        else:
            if isinstance(self.max_lr, list):
                lr = [interp.interpolate_single(max_lr, min_lr, interpolation_number, how=self.interpolate) for max_lr, min_lr in zip(self.max_lr, self.min_lr)]
            else:
                lr = interp.interpolate_single(self.max_lr, self.min_lr, interpolation_number, how=self.interpolate)

        self.set_lr(lr)

    def set_lr(self, lr):
        """ Set a learning rate for the optimizer """
        if isinstance(lr, list):
            for group_lr, param_group in zip(lr, self.optimizer.param_groups):
                param_group['lr'] = group_lr
        else:
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = lr


class CyclePhase(base.TrainPhase):
    """ Most generic phase of training """

    def __init__(self, optimizer_factory, max_lr, min_lr, cycles, cycle_len=1, cycle_mult=1, interpolate='linear',
                 init_lr=0, init_iter=0, freeze=False):
        self.max_lr = max_lr
        self.min_lr = min_lr

        self.cycles = cycles
        self.cycle_len = cycle_len
        self.cycle_mult = cycle_mult

        if cycle_mult > 1:
            self.epochs = self.cycle_len * (self.cycle_mult ** self.cycles - 1) // (self.cycle_mult - 1)
        else:
            self.epochs = self.cycle_mult * self.cycles

        self.interpolate = interpolate

        self.init_iter = init_iter
        self.init_lr = init_lr

        self.optimizer_factory = optimizer_factory
        self.freeze = freeze

        self._optimizer_instance = None
        self._source = None

        self.special_callback = None

    @property
    def number_of_epochs(self) -> int:
        return self.epochs

    def set_up_phase(self, training_info, model, source):
        """ Prepare the phase for learning """
        # To parameter groups handles properly filtering parameters that don't require gradient
        self._optimizer_instance = self.optimizer_factory.instantiate(model)
        self._source = source

        self.special_callback = CycleCallback(
            self._optimizer_instance,
            max_lr=self.max_lr, min_lr=self.min_lr, cycles=self.cycles,
            cycle_len=self.cycle_len, cycle_mult=self.cycle_mult, interpolate=self.interpolate,
            init_iter=self.init_iter, init_lr=self.init_lr
        )

        return self._optimizer_instance

    def epoch_info(self, training_info: TrainingInfo, global_idx: int, local_idx: int) -> EpochInfo:
        """ Create Epoch info """
        return EpochInfo(
            training_info=training_info,
            global_epoch_idx=global_idx,
            local_epoch_idx=local_idx,
            batches_per_epoch=self._source.train_iterations_per_epoch(),
            optimizer=self._optimizer_instance,
            # Add special callback for this epoch
            callbacks=[self.special_callback] + training_info.callbacks
        )

    def execute_epoch(self, epoch_info, learner):
        """ Prepare the phase for learning """
        learner.run_epoch(epoch_info, self._source)


def create(optimizer, max_lr, min_lr, cycles, cycle_len=1, cycle_mult=1, interpolate='linear', init_lr=0, init_iter=0):
    """ Vel creation function """
    return CyclePhase(
        max_lr=max_lr,
        min_lr=min_lr,
        cycles=cycles,
        cycle_len=cycle_len,
        cycle_mult=cycle_mult,
        interpolate=interpolate,
        optimizer_factory=optimizer,
        init_lr=init_lr,
        init_iter=init_iter,
    )



import numpy as np
import random
import torch


def set_seed(seed: int):
    """ Set random seed for python, numpy and pytorch RNGs """
    random.seed(seed)
    np.random.seed(seed)
    torch.random.manual_seed(seed)



import random
import numpy as np

import vel.api.data as data


class RandomHorizontalFlip(data.Augmentation):
    """ Apply a horizontal flip randomly to input images """

    def __init__(self, p=0.5, mode='x', tags=None):
        super().__init__(mode, tags)
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be flipped.

        Returns:
            PIL Image: Randomly flipped image.
        """
        if random.random() < self.p:
            return np.fliplr(img).copy()
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


def create(p=0.5):
    return RandomHorizontalFlip(p)


import numpy as np

from abc import ABC, abstractmethod


class AbstractEnvRunner(ABC):
    def __init__(self, *, env, model, nsteps):
        self.env = env
        self.model = model
        self.nenv = nenv = env.num_envs if hasattr(env, 'num_envs') else 1
        self.batch_ob_shape = (nenv*nsteps,) + env.observation_space.shape
        self.obs = np.zeros((nenv,) + env.observation_space.shape, dtype=env.observation_space.dtype.name)
        self.obs[:] = env.reset()
        self.nsteps = nsteps
        self.states = model.initial_state
        self.dones = [False for _ in range(nenv)]

    @abstractmethod
    def run(self):
        raise NotImplementedError



class Runner(AbstractEnvRunner):

    def __init__(self, env, model, nsteps, nstack):
        super().__init__(env=env, model=model, nsteps=nsteps)
        self.nstack = nstack
        nh, nw, nc = env.observation_space.shape
        self.nc = nc  # nc = 1 for atari, but just in case
        self.nact = env.action_space.n
        nenv = self.nenv
        self.nbatch = nenv * nsteps
        self.batch_ob_shape = (nenv*(nsteps+1), nh, nw, nc*nstack)
        self.obs = np.zeros((nenv, nh, nw, nc * nstack), dtype=np.uint8)
        obs = env.reset()
        self.raw_obs = obs.copy()
        self.update_obs(obs)

    def update_obs(self, obs, dones=None):
        # self.obs = obs
        if dones is not None:
            self.obs *= (1 - dones.astype(np.uint8))[:, None, None, None]
        self.obs = np.roll(self.obs, shift=-self.nc, axis=3)
        self.obs[:, :, :, -self.nc:] = obs[:, :, :, :]

    def run(self):
        enc_obs = np.split(self.obs, self.nstack, axis=3)  # so now list of obs steps
        mb_obs, mb_actions, mb_mus, mb_dones, mb_rewards = [], [], [], [], []
        mb_raw_obs = []

        epinfos = []

        for _ in range(self.nsteps):
            actions, mus, states = self.model.step(self.obs, S=self.states, M=self.dones)

            mb_obs.append(np.copy(self.obs))
            mb_actions.append(actions)
            mb_mus.append(mus)
            mb_dones.append(self.dones)
            obs, rewards, dones, infos = self.env.step(actions)

            for info in infos:
                maybeepinfo = info.get('episode')
                if maybeepinfo: epinfos.append(maybeepinfo)

            self.raw_obs = obs.copy()
            # states information for statefull models like LSTM
            self.states = states
            self.dones = dones
            self.update_obs(obs, dones)
            mb_rewards.append(rewards)
            enc_obs.append(obs)

            mb_raw_obs.append(self.raw_obs)

        mb_obs.append(np.copy(self.obs))
        mb_dones.append(self.dones)

        enc_obs = np.asarray(enc_obs, dtype=np.uint8).swapaxes(1, 0)
        mb_obs = np.asarray(mb_obs, dtype=np.uint8).swapaxes(1, 0)
        mb_actions = np.asarray(mb_actions, dtype=np.int32).swapaxes(1, 0)
        mb_rewards = np.asarray(mb_rewards, dtype=np.float32).swapaxes(1, 0)
        mb_mus = np.asarray(mb_mus, dtype=np.float32).swapaxes(1, 0)

        mb_raw_obs = np.asarray(mb_raw_obs, dtype=np.uint8).swapaxes(1, 0)

        mb_dones = np.asarray(mb_dones, dtype=np.bool).swapaxes(1, 0)

        mb_masks = mb_dones # Used for statefull models like LSTM's to mask state when done
        mb_dones = mb_dones[:, 1:] # Used for calculating returns. The dones array is now aligned with rewards

        # shapes are now [nenv, nsteps, []]
        # When pulling from buffer, arrays will now be reshaped in place, preventing a deep copy.

        return enc_obs, mb_obs, mb_actions, mb_rewards, mb_mus, mb_dones, mb_masks, epinfos, mb_raw_obs




"""
Code based on:
https://github.com/fastai/fastai/blob/master/fastai/transforms.py
"""
import PIL.Image as Image

import vel.api.data as data


class ScaleMinSize(data.Augmentation):
    """ Scales the image so that the smallest axis is of 'size'. """
    def __init__(self, size, mode='x', tags=None):
        super().__init__(mode, tags)
        self.size = size

    def __call__(self, x_data):
        return data.scale_min(x_data, self.size, Image.BILINEAR)


def create(size, mode='x', tags=None):
    return ScaleMinSize(size, mode, tags)



import attr
import itertools as it


@attr.s(auto_attribs=True)
class VisdomSettings:
    """ Settings for connecting to the visdom server """
    stream_lr: bool = False
    server: str = 'http://localhost'
    endpoint: str = 'events'
    port: int = 8097


def _column_original_name(name):
    """ Return original name of the metric """
    if ':' in name:
        return name.split(':')[-1]
    else:
        return name


def visdom_push_metrics(vis, metrics):
    """ Push metrics to visdom """
    visdom_send_metrics(vis, metrics, 'replace')


def visdom_send_metrics(vis, metrics, update='replace'):
    """ Send set of metrics to visdom """
    visited = {}

    sorted_metrics = sorted(metrics.columns, key=_column_original_name)
    for metric_basename, metric_list in it.groupby(sorted_metrics, key=_column_original_name):
        metric_list = list(metric_list)

        for metric in metric_list:
            if vis.win_exists(metric_basename) and (not visited.get(metric, False)):
                update = update
            elif not vis.win_exists(metric_basename):
                update = None
            else:
                update = 'append'

            vis.line(
                metrics[metric].values,
                metrics.index.values,
                win=metric_basename,
                name=metric,
                opts={
                    'title': metric_basename,
                    'showlegend': True
                },
                update=update
            )

            if metric_basename != metric and len(metric_list) > 1:
                if vis.win_exists(metric):
                    update = update
                else:
                    update = None

                vis.line(
                    metrics[metric].values,
                    metrics.index.values,
                    win=metric,
                    name=metric,
                    opts={
                        'title': metric,
                        'showlegend': True
                    },
                    update=update
                )


def visdom_append_metrics(vis, metrics, first_epoch=False):
    """ Append metrics to visdom """
    visited = {}

    sorted_metrics = sorted(metrics.columns, key=_column_original_name)
    for metric_basename, metric_list in it.groupby(sorted_metrics, key=_column_original_name):
        metric_list = list(metric_list)

        for metric in metric_list:
            if vis.win_exists(metric_basename) and (not visited.get(metric, False)) and first_epoch:
                update = 'replace'
            elif not vis.win_exists(metric_basename):
                update = None
            else:
                update = 'append'

            vis.line(
                metrics[metric].values,
                metrics.index.values,
                win=metric_basename,
                name=metric,
                opts={
                    'title': metric_basename,
                    'showlegend': True
                },
                update=update
            )

            if metric_basename != metric and len(metric_list) > 1:
                if vis.win_exists(metric) and first_epoch:
                    update = 'replace'
                elif not vis.win_exists(metric):
                    update = None
                else:
                    update = 'append'

                vis.line(
                    metrics[metric].values,
                    metrics.index.values,
                    win=metric,
                    name=metric,
                    opts={
                        'title': metric,
                        'showlegend': True
                    },
                    update=update
                )




import gym


class ClipEpisodeLengthWrapper(gym.Wrapper):
    """ Env wrapper that clips number of frames an episode can last """
    def __init__(self, env, max_episode_length):
        super().__init__(env)

        self.max_episode_length = max_episode_length
        self.current_episode_length = 0

    def reset(self, **kwargs):
        self.current_episode_length = 0
        return self.env.reset(**kwargs)

    def step(self, action):
        self.current_episode_length += 1
        ob, reward, done, info = self.env.step(action)

        if self.current_episode_length > self.max_episode_length:
            done = True
            info['clipped_length'] = True

        return ob, reward, done, info



from torchvision import datasets

from vel.api.base import TrainingData

from vel.augmentations.normalize import Normalize
from vel.augmentations.to_tensor import ToTensor
from vel.augmentations.to_array import ToArray


def create(batch_size, model_config, normalize=True, num_workers=0, augmentations=None):
    """
    Create a CIFAR10 dataset, normalized.
    Augmentations are the same as in the literature benchmarking CIFAR performance.
    """
    path = model_config.data_dir('cifar10')

    train_dataset = datasets.CIFAR10(path, train=True, download=True)
    test_dataset = datasets.CIFAR10(path, train=False, download=True)

    augmentations = [ToArray()] + (augmentations if augmentations is not None else [])
    
    if normalize:
        train_data = train_dataset.train_data
        mean_value = (train_data / 255).mean(axis=(0, 1, 2))
        std_value = (train_data / 255).std(axis=(0, 1, 2))

        augmentations.append(Normalize(mean=mean_value, std=std_value, tags=['train', 'val']))

    augmentations.append(ToTensor())

    return TrainingData(
        train_dataset,
        test_dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        augmentations=augmentations
    )



import torch
import numpy as np

from vel.rl.api.base import EnvRollerBase, EnvRollerFactory
from vel.rl.api import Trajectories


class StepEnvRoller(EnvRollerBase):
    """
    Class calculating env rollouts.
    Idea behind this class is to store as much as we can as pytorch tensors to minimize tensor copying.
    """

    def __init__(self, environment, device, number_of_steps, discount_factor, gae_lambda=1.0):
        self._environment = environment
        self.device = device
        self.number_of_steps = number_of_steps
        self.discount_factor = discount_factor
        self.gae_lambda = gae_lambda

        # Initial observation
        self.last_observation = self._to_tensor(self.environment.reset())

        # Relevant for RNN policies
        self.hidden_state = None

    @property
    def environment(self):
        """ Return environment of this env roller """
        return self._environment

    def _to_tensor(self, numpy_array):
        """ Convert numpy array to a tensor """
        return torch.from_numpy(numpy_array).to(self.device)

    @torch.no_grad()
    def rollout(self, batch_info, model):
        """ Calculate env rollout """
        observation_accumulator = []  # Device tensors
        action_accumulator = []  # Device tensors
        value_accumulator = []  # Device tensors
        dones_accumulator = []  # Device tensors
        rewards_accumulator = []  # Device tensors
        episode_information = []  # Python objects
        logprobs_accumulator = []  # Device tensors

        if self.hidden_state is None and model.is_recurrent:
            self.hidden_state = torch.zeros(
                (self.last_observation.size(0), model.state_dim),
                device=self.device,
                dtype=torch.float32
            )

        # Remember rollout initial state, we'll use that for learning as well
        initial_hidden_state = self.hidden_state

        for step_idx in range(self.number_of_steps):
            if model.is_recurrent:
                step = model.step(self.last_observation, state=self.hidden_state)
                self.hidden_state = step['state']
            else:
                step = model.step(self.last_observation)

            actions, values, logprobs = step['actions'], step['values'], step['logprobs']

            observation_accumulator.append(self.last_observation)
            action_accumulator.append(actions)
            value_accumulator.append(values)
            logprobs_accumulator.append(logprobs)

            actions_numpy = actions.detach().cpu().numpy()
            new_obs, new_rewards, new_dones, new_infos = self.environment.step(actions_numpy)

            # Done is flagged true when the episode has ended AND the frame we see is already a first frame from the
            # next episode

            dones_tensor = self._to_tensor(new_dones.astype(np.float32))
            dones_accumulator.append(dones_tensor)

            self.last_observation = self._to_tensor(new_obs[:])

            if model.is_recurrent:
                # Zero out state in environments that have finished
                self.hidden_state = self.hidden_state * (1.0 - dones_tensor.unsqueeze(-1))

            rewards_accumulator.append(self._to_tensor(new_rewards.astype(np.float32)))

            episode_information.append(new_infos)

        if model.is_recurrent:
            final_values = model.value(self.last_observation, state=self.hidden_state)
        else:
            final_values = model.value(self.last_observation)

        observations_buffer = torch.stack(observation_accumulator)
        rewards_buffer = torch.stack(rewards_accumulator)
        actions_buffer = torch.stack(action_accumulator)  # Actions may have various different dtypes
        values_buffer = torch.stack(value_accumulator)
        dones_buffer = torch.stack(dones_accumulator)
        logprobs_buffer = torch.stack(logprobs_accumulator)

        # Generalized Advantage Estimation
        # https://arxiv.org/abs/1506.02438
        advantages = self.discount_bootstrap_gae(
            rewards_buffer, dones_buffer, values_buffer, final_values,
            self.discount_factor, self.gae_lambda
        )

        returns = advantages + values_buffer

        return Trajectories(
            num_steps=advantages.size(0),
            num_envs=advantages.size(1),
            environment_information=episode_information,
            transition_tensors={
                'observations': observations_buffer,
                'estimated_returns': returns,
                'dones': dones_buffer,
                'actions': actions_buffer,
                'estimated_values': values_buffer,
                'estimated_advantages': advantages,
                'action:logprobs': logprobs_buffer,
            },
            rollout_tensors={
                'initial_hidden_state': initial_hidden_state,
                'final_estimated_values': final_values
            }
        )

    def discount_bootstrap(self, rewards_buffer, dones_buffer, final_values, discount_factor):
        """ Calculate state values bootstrapping off the following state values """
        true_value_buffer = torch.zeros_like(rewards_buffer)
        dones_buffer = dones_buffer

        # discount/bootstrap off value fn
        current_value = final_values

        for i in reversed(range(self.number_of_steps)):
            current_value = rewards_buffer[i] + discount_factor * current_value * (1.0 - dones_buffer[i])
            true_value_buffer[i] = current_value

        return true_value_buffer

    def discount_bootstrap_gae(self, rewards_buffer, dones_buffer, final_values, last_values_buffer,
                               discount_factor, gae_lambda):
        """ Calculate state values bootstrapping off the following state values - Generalized Advantage Estimation """
        advantage_buffer = torch.zeros_like(rewards_buffer)
        dones_buffer = dones_buffer

        # Accmulate sums
        sum_accumulator = 0

        for i in reversed(range(self.number_of_steps)):
            if i == self.number_of_steps - 1:
                next_value = last_values_buffer
            else:
                next_value = final_values[i + 1]

            bellman_delta = (
                    rewards_buffer[i] + discount_factor * next_value * (1.0 - dones_buffer[i]) - final_values[i]
            )

            advantage_buffer[i] = sum_accumulator = (
                bellman_delta + discount_factor * gae_lambda * sum_accumulator * (1.0 - dones_buffer[i])
            )

        return advantage_buffer


class StepEnvRollerFactory(EnvRollerFactory):
    """ Factory for the StepEnvRoller """
    def __init__(self, number_of_steps, gae_lambda=1.0):
        self.gae_lambda = gae_lambda
        self.number_of_steps = number_of_steps

    def instantiate(self, environment, device, settings):
        return StepEnvRoller(
            environment=environment,
            device=device,
            number_of_steps=self.number_of_steps,
            discount_factor=settings.discount_factor,
            gae_lambda=self.gae_lambda
        )


def create(number_of_steps, gae_lambda=1.0):
    return StepEnvRollerFactory(number_of_steps=number_of_steps, gae_lambda=gae_lambda)
